@@ -6,13 +6,13 @@ use Net::Stomp::Frame;
 use Carp qw(longmess);
 use base 'Class::Accessor::Fast';
 use Net::Stomp::StupidLogger;
-our $VERSION = '0.49';
+our $VERSION = '0.56';
 
 __PACKAGE__->mk_accessors( qw(
     current_host failover hostname hosts port select serial session_id socket ssl
-    ssl_options subscriptions _connect_headers bufsize
+    ssl_options socket_options subscriptions _connect_headers bufsize
     reconnect_on_fork logger connect_delay
-    reconnect_attempts initial_reconnect_attempts
+    reconnect_attempts initial_reconnect_attempts timeout receipt_timeout
 ) );
 
 sub _logconfess {
@@ -36,6 +36,7 @@ sub new {
     $self->reconnect_on_fork(1) unless defined $self->reconnect_on_fork;
     $self->reconnect_attempts(0) unless defined $self->reconnect_attempts;
     $self->initial_reconnect_attempts(1) unless defined $self->initial_reconnect_attempts;
+    $self->socket_options({}) unless defined $self->socket_options;
 
     $self->logger(Net::Stomp::StupidLogger->new())
         unless $self->logger;
@@ -136,12 +137,18 @@ sub _get_socket {
     my ($self) = @_;
     my $socket;
 
+    my $timeout = $self->timeout;
+    $timeout = 5 unless defined $timeout;
+
     my %sockopts = (
+        Timeout  => $timeout,
+        %{ $self->socket_options },
         PeerAddr => $self->hostname,
         PeerPort => $self->port,
         Proto    => 'tcp',
-        Timeout  => 5
     );
+    my $keep_alive = delete $sockopts{keep_alive};
+
     if ( $self->ssl ) {
         eval { require IO::Socket::SSL };
         $self->_logdie(
@@ -155,6 +162,15 @@ sub _get_socket {
         $socket = $socket_class->new(%sockopts);
         binmode($socket) if $socket;
     }
+    if ($keep_alive) {
+        require Socket;
+        if (Socket->can('SO_KEEPALIVE')) {
+            $socket->setsockopt(Socket::SOL_SOCKET(),Socket::SO_KEEPALIVE(),1);
+        }
+        else {
+            $self->logger->warn(q{TCP keep-alive was requested, but the Socket module does not export the SO_KEEPALIVE constant, so we couldn't enable it});
+        }
+    }
 
     return $socket;
 }
@@ -189,6 +205,7 @@ sub disconnect {
     my $frame = Net::Stomp::Frame->new( { command => 'DISCONNECT' } );
     $self->send_frame($frame);
     $self->_close_socket;
+    return 1;
 }
 
 sub _reconnect {
@@ -219,42 +236,37 @@ sub can_read {
     }
 
     $conf ||= {};
-    my $timeout = exists $conf->{timeout} ? $conf->{timeout} : undef;
+    my $timeout = exists $conf->{timeout} ? $conf->{timeout} : $self->timeout;
     return $self->select->can_read($timeout) || 0;
 }
 
 sub send {
     my ( $self, $conf ) = @_;
+    $conf = { %$conf };
     my $body = $conf->{body};
     delete $conf->{body};
     my $frame = Net::Stomp::Frame->new(
         { command => 'SEND', headers => $conf, body => $body } );
     $self->send_frame($frame);
+    return 1;
 }
 
-sub send_transactional {
+sub send_with_receipt {
     my ( $self, $conf ) = @_;
-    my $body = $conf->{body};
-    delete $conf->{body};
-
-    # begin the transaction
-    my $transaction_id = $self->_get_next_transaction;
-    my $begin_frame
-        = Net::Stomp::Frame->new(
-        { command => 'BEGIN', headers => { transaction => $transaction_id } }
-        );
-    $self->send_frame($begin_frame);
+    $conf = { %$conf };
 
     # send the message
     my $receipt_id = $self->_get_next_transaction;
     $conf->{receipt} = $receipt_id;
-    $conf->{transaction} = $transaction_id;
-    my $message_frame = Net::Stomp::Frame->new(
-        { command => 'SEND', headers => $conf, body => $body } );
-    $self->send_frame($message_frame);
+    my $receipt_timeout = exists $conf->{timeout} ? delete $conf->{timeout} : $self->receipt_timeout;
+    $self->send($conf);
 
     # check the receipt
-    my $receipt_frame = $self->receive_frame;
+    my $receipt_frame = $self->receive_frame({
+        ( defined $receipt_timeout ?
+              ( timeout => $receipt_timeout )
+              : () ),
+    });
 
     if (@_ > 2) {
         $_[2] = $receipt_frame;
@@ -264,7 +276,33 @@ sub send_transactional {
         && $receipt_frame->command eq 'RECEIPT'
         && $receipt_frame->headers->{'receipt-id'} eq $receipt_id )
     {
+        return 1;
+    } else {
+        return 0;
+    }
+}
 
+sub send_transactional {
+    my ( $self, $conf ) = @_;
+
+    $conf = { %$conf };
+    # begin the transaction
+    my $transaction_id = $self->_get_next_transaction;
+    my $begin_frame
+        = Net::Stomp::Frame->new(
+        { command => 'BEGIN', headers => { transaction => $transaction_id } }
+        );
+    $self->send_frame($begin_frame);
+
+    $conf->{transaction} = $transaction_id;
+    my $receipt_frame;
+    my $ret = $self->send_with_receipt($conf,$receipt_frame);
+
+    if (@_ > 2) {
+        $_[2] = $receipt_frame;
+    }
+
+    if ( $ret ) {
         # success, commit the transaction
         my $frame_commit = Net::Stomp::Frame->new(
             {   command => 'COMMIT',
@@ -272,9 +310,7 @@ sub send_transactional {
             }
         );
         $self->send_frame($frame_commit);
-        return 1;
     } else {
-
         # some failure, abort transaction
         my $frame_abort = Net::Stomp::Frame->new(
             {   command => 'ABORT',
@@ -282,8 +318,8 @@ sub send_transactional {
             }
         );
         $self->send_frame($frame_abort);
-        return 0;
     }
+    return $ret;
 }
 
 sub _sub_key {
@@ -300,6 +336,7 @@ sub subscribe {
     $self->send_frame($frame);
     my $subs = $self->subscriptions;
     $subs->{_sub_key($conf)} = $conf;
+    return 1;
 }
 
 sub unsubscribe {
@@ -308,16 +345,19 @@ sub unsubscribe {
         { command => 'UNSUBSCRIBE', headers => $conf } );
     $self->send_frame($frame);
     my $subs = $self->subscriptions;
-    delete $subs->{_sub_key($conf)}
+    delete $subs->{_sub_key($conf)};
+    return 1;
 }
 
 sub ack {
     my ( $self, $conf ) = @_;
+    $conf = { %$conf };
     my $id    = $conf->{frame}->headers->{'message-id'};
     delete $conf->{frame};
     my $frame = Net::Stomp::Frame->new(
         { command => 'ACK', headers => { 'message-id' => $id, %$conf } } );
     $self->send_frame($frame);
+    return 1;
 }
 
 sub send_frame {
@@ -346,6 +386,7 @@ sub send_frame {
         $self->_reconnect;
         $self->send_frame($frame);
     }
+    return;
 }
 
 sub _read_data {
@@ -377,6 +418,7 @@ sub _read_data {
 sub _read_headers {
     my ($self) = @_;
 
+    return 1 if $self->{_headers};
     if ($self->{_framebuf} =~ s/^\n*([^\n].*?)\n\n//s) {
         $self->{_framebuf_changed} = 1;
         my $raw_headers = $1;
@@ -398,7 +440,7 @@ sub _read_body {
 
     my $h = $self->{_headers};
     if ($h->{'content-length'}) {
-        if (length($self->{_framebuf}) >= $h->{'content-length'}) {
+        if (length($self->{_framebuf}) > $h->{'content-length'}) {
             $self->{_framebuf_changed} = 1;
             my $body = substr($self->{_framebuf},
                               0,
@@ -450,7 +492,7 @@ sub _connected {
 sub receive_frame {
     my ($self, $conf) = @_;
 
-    my $timeout = exists $conf->{timeout} ? $conf->{timeout} : undef;
+    my $timeout = exists $conf->{timeout} ? $conf->{timeout} :  $self->timeout;
 
     unless (defined $self->_connected) {
         $self->_reconnect;
@@ -652,6 +694,18 @@ port here. If you modify this value during the lifetime of the
 object, the new value will be used for the subsequent reconnect
 attempts.
 
+=head2 C<socket_options>
+
+Optional hashref, it will be passed to the L<IO::Socket::IP>,
+L<IO::Socket::SSL>, or L<IO::Socket::INET> constructor every time we
+need to get a socket.
+
+In addition to the various options supported by those classes, you can
+set C<keep_alive> to a true value, which will enable TCP-level
+keep-alive on the socket (see L<the TCP Keepalive
+HOWTO|http://www.tldp.org/HOWTO/html_single/TCP-Keepalive-HOWTO/> for
+some information on that feature).
+
 =head2 C<ssl>
 
 Boolean, defaults to false, whether we should use SSL to talk to the
@@ -661,9 +715,9 @@ attempts.
 
 =head2 C<ssl_options>
 
-Options to pass to L<< /C<IO::Socket::SSL> >> when connecting via SSL
-to the single broker. If you modify this value during the lifetime of
-the object, the new value will be used for the subsequent reconnect
+Options to pass to L<IO::Socket::SSL> when connecting via SSL to the
+single broker. If you modify this value during the lifetime of the
+object, the new value will be used for the subsequent reconnect
 attempts.
 
 =head2 C<failover>
@@ -725,6 +779,18 @@ sleep of L<< /C<connect_delay> >> seconds.
 Integer, defaults to 5. How many seconds to sleep between connection
 attempts to brokers.
 
+=head2 C<timeout>
+
+Integer, in seconds, defaults to C<undef>. The default timeout for
+read operations. C<undef> means "wait forever".
+
+=head2 C<receipt_timeout>
+
+Integer, in seconds, defaults to C<undef>. The default timeout while
+waiting for a receipt (in L<< /C<send_with_receipt> >> and L<<
+/C<send_transactional> >>). If C<undef>, the global L<< /C<timeout> >>
+is used.
+
 =head1 METHODS
 
 =head2 C<connect>
@@ -751,10 +817,60 @@ It's probably a good idea to pass a C<content-length> corresponding to
 the byte length of the C<body>; this is necessary if the C<body>
 contains a byte 0.
 
+Always returns a true value. It automatically reconnects if writing to
+the socket fails.
+
+=head2 C<send_with_receipt>
+
+This sends a message asking for a receipt, and returns false if the
+receipt of the message is not acknowledged by the server:
+
+  $stomp->send_with_receipt(
+      { destination => '/queue/foo', body => 'test message' }
+  ) or die "Couldn't send the message!";
+
+If using ActiveMQ, you might also want to make the message persistent:
+
+  $stomp->send_transactional(
+      { destination => '/queue/foo', body => 'test message', persistent => 'true' }
+  ) or die "Couldn't send the message!";
+
+The actual frame sequence for a successful sending is:
+
+  -> SEND
+  <- RECEIPT
+
+The actual frame sequence for a failed sending is:
+
+  -> SEND
+  <- anything but RECEIPT
+
+If you are using this connection only to send (i.e. you've never
+called L<< /C<subscribe> >>), the only thing that could be received
+instead of a C<RECEIPT> is an C<ERROR> frame, but if you subscribed,
+the broker may well send a C<MESSAGE> before sending the
+C<RECEIPT>. B<DO NOT> use this method on a connection used for
+receiving.
+
+If you want to see the C<RECEIPT> or C<ERROR> frame, pass a scalar as
+a second parameter to the method, and it will be set to the received
+frame:
+
+  my $success = $stomp->send_transactional(
+      { destination => '/queue/foo', body => 'test message' },
+      $received_frame,
+  );
+  if (not $success) { warn $received_frame->as_string }
+
+You can specify a C<timeout> in the parametrs, just like for L<<
+/C<received_frame> >>. This function will wait for that timeout, or
+for L<< /C<receipt_timeout> >>, or for L<< /C<timeout> >>, whichever
+is defined, or forever, if none is defined.
+
 =head2 C<send_transactional>
 
-This sends a message in transactional mode and fails if the receipt of the
-message is not acknowledged by the server:
+This sends a message in transactional mode and returns false if the
+receipt of the message is not acknowledged by the server:
 
   $stomp->send_transactional(
       { destination => '/queue/foo', body => 'test message' }
@@ -766,6 +882,9 @@ If using ActiveMQ, you might also want to make the message persistent:
       { destination => '/queue/foo', body => 'test message', persistent => 'true' }
   ) or die "Couldn't send the message!";
 
+C<send_transactional> just wraps C<send_with_receipt> in a STOMP
+transaction.
+
 The actual frame sequence for a successful sending is:
 
   -> BEGIN
@@ -797,6 +916,11 @@ frame:
   );
   if (not $success) { warn $received_frame->as_string }
 
+You can specify a C<timeout> in the parametrs, just like for L<<
+/C<received_frame> >>. This function will wait for that timeout, or
+for L<< /C<receipt_timeout> >>, or for L<< /C<timeout> >>, whichever
+is defined, or forever, if none is defined.
+
 =head2 C<disconnect>
 
 This disconnects from the Stomp server:
@@ -807,11 +931,15 @@ If you call any other method after this, a new connection will be
 established automatically (to the next failover host, if there's more
 than one).
 
+Always returns a true value.
+
 =head2 C<subscribe>
 
 This subscribes you to a queue or topic. You must pass in a
 C<destination>.
 
+Always returns a true value.
+
 The acknowledge mode (header C<ack>) defaults to C<auto>, which means
 that frames will be considered delivered after they have been sent to
 a client. The other option is C<client>, which means that messages
@@ -828,7 +956,7 @@ Other options:
 =item C<selector>
 
 Specifies a JMS Selector using SQL 92 syntax as specified in the JMS
-1.1 specificiation. This allows a filter to be applied to each message
+1.1 specification. This allows a filter to be applied to each message
 as part of the subscription.
 
 =item C<id>
@@ -901,6 +1029,8 @@ C<destination> or an C<id>:
 
   $stomp->unsubcribe({ destination => '/queue/foo' });
 
+Always returns a true value.
+
 =head2 C<receive_frame>
 
 This blocks and returns you the next Stomp frame, or C<undef> if there
@@ -909,8 +1039,9 @@ was a connection problem.
   my $frame = $stomp->receive_frame;
   warn $frame->body; # do something here
 
-By default this method will block until a frame can be returned. If you wish to
-wait for a specified time pass a C<timeout> argument:
+By default this method will block until a frame can be returned, or
+for however long the L</timeout> attribue says. If you wish to wait
+for a specified time pass a C<timeout> argument:
 
   # Wait half a second for a frame, else return undef
   $stomp->receive_frame({ timeout => 0.5 })
@@ -923,8 +1054,9 @@ STOMP server. Optionally takes a timeout in seconds:
   my $can_read = $stomp->can_read;
   my $can_read = $stomp->can_read({ timeout => '0.1' });
 
-C<undef> says block until something can be read, C<0> says to poll and return
-immediately.
+C<undef> says block until something can be read, C<0> says to poll and
+return immediately. This method ignores the value of the L</timeout>
+attribute.
 
 =head2 C<ack>
 
@@ -933,6 +1065,8 @@ all frames before it> (if you are using client acknowledgements):
 
   $stomp->ack( { frame => $frame } );
 
+Always returns a true value.
+
 =head2 C<send_frame>
 
 If this module does not provide enough help for sending frames, you
@@ -949,6 +1083,8 @@ the connection breaks (limited by L<< /C<reconnect_attempts> >>). If
 no connection can be established, and L<< /C<reconnect_attempts> >> is
 not 0, this method will C<die>.
 
+Always returns an empty list.
+
 =head1 SEE ALSO
 
 L<Net::Stomp::Frame>.
